##############################################################################
# Exam Class, part of the Exam Product for Zope
#
# Copyright (C) 2001 Dr John A.R. Williams
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
#
##############################################################################

"""Exam Product"""

from OFS.Folder import Folder
import Products
import Globals
from Globals import PersistentMapping, HTMLFile, HTML
from DateTime import DateTime
from whrandom import randint
import ExamQuestion
import MoreExamQuestions

manage_addExamForm=HTMLFile('addExamForm',globals())

def manage_addExam(self, id, title, REQUEST=None):
    """Add a Tutorial object
    """
    id=str(id)
    title=str(title)
    c=Exam(id, title)
    self._setObject(id, c)
    if REQUEST is not None:
        return self.manage_main(self, REQUEST)

class Exam( Folder ):
  """Exam Base object"""
  
  meta_type = "Exam"
  icon="misc_/Exam/exam"

  meta_types=({'permission': 'Manage Tutors Information',
               'name': 'Exam Question',
               'product': 'Exam',
               'action': 'manage_addExamQuestionForm'},
              {'permission': 'Manage Tutors Information',
               'name': 'Exam MCQ',
               'product': 'Exam',
               'action': 'manage_addExamMCQForm'},
              {'permission': 'Manage Tutors Information',
               'name': 'Exam Numeric Question',
               'product': 'Exam',
               'action': 'manage_addExamNumericQForm'},
              {'permission': 'Manage Tutors Information', # extra
               'name': 'Exam MAQ',
               'product': 'Exam',
               'action': 'manage_addExamMAQForm'},
              )

  default_settings={
    'feedback_date':None,
    'start_date':None,
    'end_date':None,
    'randomise_questions':1,
    'set_no_questions':None,
    'set_weighting':None,
    'show_weighting':0,
    'deadline':None,
    'strict_deadline':0,
    'timelimit': None,
    'strict_timelimit': 0,
    'classlist':None,
    'userid':None,
    }
  
  # most methods require 'Access contents information'. Students should not
  # be granted this except through proxy
  __ac_permissions__=(
    ('Manage Tutors Information',
     ('manage_submissions','manage_editSubmissionForm',
      'manage_delSubmission', 'manage_changeSubmission',
      'manage_addExamQuestionForm','manage_addExamQuestion',
      'manage_addExamMCQForm', 'manage_addExamMCQ',
      'manage_addExamMAQForm', 'manage_addExamMAQ', # extra
      'manage_perm',),),
    ('View Tutors Information',
     ('view_all_questions','classIds','assessedIds','accessedIds',
      'submittedIds','all_questions','question_statistics','report_html',
      'collusion_set'),),
    ('Access contents information',
     ('getSetting','getUserid','timetaken','getSubmissionRecord',
      'exam_submit','exam_form',
      'exam_feedback','noQuestions',
      'getQuestion','newQuestionSet','assessmentStatus',
      'getQuestions','exam_init','exam_form',
      'exam_submit','exam_feedback','histogram','mark','raw_mark',
      'question_statistics','average_time','summary_details'
      ),),
    ('View', ('accessCheck','submitCheck','feedbackCheck','index_html',
              'feedback_html','exam_html','submit_html'),),
    #('Default permission', ('index_html',)),
    )
 
  manage_options=Folder.manage_options+ (
      {'label':'View Questions','action':'manage_view',},
      {'label':'Submissions','action':'manage_submissions',},
      )

  def __init__(self, id,title):
    self.id = id
    self.title=title
    self._submissions=PersistentMapping()

  manage_addExamQuestionForm=HTMLFile('addExamQuestionForm',globals())

  def manage_addExamQuestion(self, id, REQUEST=None):
      """Add an Exam Question"""
      id=str(id)
      c=ExamQuestion.ExamQuestion(id)
      self._setObject(id, c)
      if REQUEST is not None:
          return self.manage_main(self, REQUEST)

  manage_addExamMCQForm=HTMLFile('addExamMCQForm',globals())

  def manage_addExamMCQ(self, id, REQUEST=None):
      """Add an MCQ"""
      id=str(id)
      c=ExamQuestion.ExamMCQ(id)
      self._setObject(id, c)
      if REQUEST is not None:
          return self.manage_main(self, REQUEST)

  manage_addExamNumericQForm=HTMLFile('addExamNumericQForm',globals())

  def manage_addExamNumericQ(self, id, REQUEST=None):
      """Add an Exam Numeric Question"""
      id=str(id)
      c=ExamQuestion.ExamNumericQ(id)
      self._setObject(id, c)
      if REQUEST is not None:
          return self.manage_main(self, REQUEST)

  manage_addExamMAQForm=HTMLFile('addExamMAQForm',globals())

  def manage_addExamMAQ(self, id, REQUEST=None):
      """Add an MAQ"""
      id=str(id)
      c=MoreExamQuestions.ExamMAQ(id)
      self._setObject(id, c)
      if REQUEST is not None:
          return self.manage_main(self, REQUEST)

  def getSetting(self,name,default=None):
    """Return an aquired setting - possible callable"""
    if hasattr(self,name):
      o= getattr(self,name)
      if callable(o):
        return o(self)
      else:
        return o
    if default is not None: return default
    return self.default_settings[name]

  def getUserid(self):
    """Return the userid either from userid method or AUTHENTICATED_USER"""
    userid=self.getSetting('userid')
    if userid is None:
      userid=self.getSetting('REQUEST').AUTHENTICATED_USER.getUserName()
    return userid

  def timetaken(self,userid=None):
    """time taken for given student"""
    if userid is None: userid=self.getUserid()
    if not self._submissions.has_key(userid): return None
    sub= self._submissions[userid]
    if sub['submitted']:
      return 1440*(sub['submitted']-sub['started'])
    else:
      return 1440*(DateTime()-sub['started'])
    
  def getSubmissionRecord(self,userid=None):
    """Return a hash containing students submission information"""
    if userid is None: userid=self.getUserid()
    if not self._submissions.has_key(userid): return None
    sub= self._submissions[userid]._container.copy()
    sub['timetaken']=self.timetaken(userid)
    if sub['submitted']:
      sub['mark']=self.mark(userid)
    deadline=self.getSetting('deadline')
    sub['deadline']=deadline
    timelimit=self.getSetting('timelimit')
    if timelimit is not None:
      d=sub['started']+timelimit/1440.0
      if deadline is None or d<deadline: 
        sub['deadline']=d
    sub['status']=self.assessmentStatus(userid)
    sub['remaining']=(sub['deadline']-DateTime())*1440.0
    return sub
  
  def manage_delSubmission(self,userids,REQUEST=None):
    """Delete named submissions"""
    for q in self.all_questions():
      q.manage_delSubmission(userids)
    for userid in userids:
      if self._submissions.has_key(userid):
        del self._submissions[userid]
    if REQUEST is not None:
      return self.manage_submissions(self, REQUEST)

  def manage_changeSubmission(self,userid,ids,started=None,
                              submitted=None,REQUEST=None):
    """ Add or change a submission ids=question ids """
    sub=PersistentMapping()
    if started is None: started=DateTime()
    sub['started']=started
    sub['questionids']=ids
    sub['submitted']=None
    if submitted is not None:
      try:
        sub['submitted']=DateTime(submitted)
      except:
        pass
    self._submissions[userid]=sub
    if REQUEST is not None:
      return self.manage_submissions(self, REQUEST)

  manage_view=HTMLFile("examView", globals())

  def view_all_questions(self):
    qs=self.all_questions()
    f='<ol>'
    for q in qs:
      f=f+('\n<li><p><strong>'+ q.title_or_id()+
           '</strong> [%2.2f] </p>' % q.weighting(userid=None) +
           q.feedback(None))
    return f+'</ol>\n'

  manage_submissions=HTMLFile("examSubmissions", globals())

  manage_editSubmissionForm=HTMLFile("examEditSubmission", globals())

  def _clearSubmissions(self):
    for q in self.all_questions():
      q.manage_delSubmission()
    self._submissions.clear()
    
  def all_questions(self):
    """ return the list of all available questions by order of id"""
    #should perhaps use some form of superValues or something
    q=[]
    for obj in self.objectValues():
      if hasattr(obj,'isQuestion') and obj.isQuestion:
        q.append(obj)
    q.sort(lambda a,b: cmp(a.id,b.id))
    return q

  def noQuestions(self,userid=None):
    """return number of questions for a given student or None
    if it is not defined yet"""
    if userid is None: userid=self.getUserid()
    if self._submissions.has_key(userid):
      return len(self._submissions[userid]['questionids'])
    total=len(self.all_questions())
    n=self.getSetting('set_no_questions')
    if n:
      if n<total: return n
    else:
      if self.getSetting('set_weighting'): return None
    return total

  def getQuestion(self,id):
    return getattr(self,id)
    
  def newQuestionSet(self,userid=None):
    """ return a new question set according to the settings"""
    if userid is None: userid=self.getUserid()
    q=self.all_questions()
    # randomise if necessary
    if self.getSetting('randomise_questions'):
      newlist=[]
      while len(q)>1: 
        n=randint(0,len(q)-1)
	newlist.append(q[n])
	del q[n]
      newlist.append(q[0])
      q=newlist
    # if a fixed no of questions to be used, return only that many
    n=self.getSetting('set_no_questions')
    if n:
      if n<len(q):
        return q[:n]
      else:
        return q
    # if we have defined a particular weighting then use it
    n=self.getSetting('set_weighting')
    if n:
      w=0
      for i in range(len(q)):
        w=w+q[i].weighting()
        if w>n: return q[0:i]
    return q

  def assessmentStatus(self,userid=None):
    """check if assessment for student counts
    return a tuple of a code and long message string
    if doesn't count, otherwise return None"""
    if userid is None: userid=self.getUserid()
    if not userid in self.classIds(): return ('NA','Not Assessed')
    if not self._submissions.has_key(userid): return ('NS','No Submission')
    sub=self._submissions[userid]
    if sub['submitted']:
      d=self.getSetting('deadline')
      if d and sub['submitted']>=d: return ('MD','Missed Deadline')
      if self.getSetting('strict_timelimit'):
        d=self.getSetting('timelimit')
        if d and 1440*(sub['submitted']-sub['started'])>d:
          return ('TE','Timeout Exceeded')
    return None

  def accessCheck(self,userid=None):
    """Check to see if a given student is allowed to access this questionnaire at all"""
    if userid is None: userid=self.getUserid()
    d=self.getSetting('start_date')
    if d and d.isFuture():
        raise 'Forbidden',('Questionnaire "'+self.title_or_id()+
                           '" not available until '+d.aCommon())
    d=self.getSetting('end_date')
    if d and d.isPast():
        raise 'Forbidden',('Questionnaire "'+self.title_or_id()+
               '" not available after '+d.aCommon())
    c=self.getSetting('classlist')
    if c and (userid not in c):
      raise 'Unauthorized', (
          'You are not on the classlist for questionnaire "'+
        self.title_or_id()+'" and therefore cannot access it.')
    
  def submitCheck(self,userid=None):
    """Check to see if a student is allowed to submit answers to this exam"""
    if userid is None: userid=self.getUserid()
    self.accessCheck(userid)
    d=self. getSetting('deadline')
    if d and d.isPast():
        if (self.getSetting('strict_deadline')
            or self._submissions.has_key(userid)):
            raise 'Forbidden', (
                'The submission deadline for questionnaire "'
                +self.title_or_id()+'" was '+d.aCommon())
    if not self._submissions.has_key(userid): return
    sub=self._submissions[userid]
    if (sub.has_key('submitted') and self.getSetting('strict_timelimit')
       and  1440*(time()-sub['started'])>self.getSetting('timelimit')
      ):
      raise 'Forbidden', (
          'You are trying to submit answers to Questionnaire'+
          self.title_or_id()+
          ' outside the timelimit. You have already submitted answers on'+
          sub['submitted']
          )

  def feedbackCheck(self,userid=None):
    """check to see if student can get feedback"""
    if userid is None: userid=self.getUserid()
    self.accessCheck(userid)
    if not self._submissions.has_key(userid):
      raise 'Forbidden', (
          'You cannot get feedback since you have not submitted answers to Questionnaire "'
          +self.title_or_id()+'"')
    d=self.getSetting('feedback_date')
    if d and d.isFuture():
      raise 'Forbidden',('Feedback on "'+self.title_or_id()+
                         '" not available until '+d.aCommon())

  def classIds(self):
    """ reference to a list of all ids who should be considered for
    assessment (even if they haven't submitted yet)"""
    return self.getSetting('classlist') or self.submittedIds()

  def assessedIds(self):
    """students who have submitted a valid submission"""
    return filter(
        lambda id,self=self: self.assessmentStatus(id) is None,
        self.classIds()
        )
  
  def accessedIds(self):
    """list of the ids of all who have accessed the Exam"""
    return self._submissions.keys() or ()

  def submittedIds(self):
    """list of ids of all who have submitted (completed) the Exam"""
    return filter(
      lambda u,s=self._submissions: s[u].has_key('submitted'),
        self.accessedIds())

  def getQuestions(self,userid):
    """return the preallocated array of Exam questions for given student"""
    return filter(
      lambda obj,ids=self._submissions[userid]['questionids']:
      obj.id in ids
      ,self.all_questions())

  def exam_init(self,userid=None):
    """initialise submission record for a given userid, if necessary """
    if userid is None: userid=self.getUserid()
    self.submitCheck(userid)
    if not self._submissions.has_key(userid):
      self.manage_changeSubmission(
        userid, map ( lambda obj: obj['id'], self.newQuestionSet(userid) ))

  def exam_form(self,userid=None,action='submit_html'):
    """returns exam questions form """
    if userid is None: userid=self.getUserid()
    self.exam_init(userid)
    qs=self.getQuestions(userid)
    f=''
    f=f+'<form action="'+action+'" method="POST">\n<ol>'
    for q in qs:
      f=f+'\n<li>'
      if self.getSetting('show_weighting'):
          f=f+'[%2.2f] ' % q.weighting(userid)
      f=f+q.question(userid)+'</li>'
      
    return f+'\n<input type="submit" name="submit" value="Submit">\n</form>\n'

  def exam_submit(self,REQUEST,userid=None):
    """takes a submission and updates records"""
    if userid is None: userid=self.getUserid()
    self.submitCheck(userid)
    self._submissions[userid]['submitted']=DateTime()
    for qid in self._submissions[userid]['questionids']:
      getattr(self, qid).manage_changeSubmission(userid,REQUEST,1)
    
  def exam_feedback(self,userid=None):
    """return feedback for given student, throws exception if not allowed"""
    if userid is None: userid=self.getUserid()
    self.feedbackCheck(userid)
    qs=self.getQuestions(userid)
    f='<ol>'
    for q in qs:
        f=f+'\n<li>'
        if self.getSetting('show_weighting'):
            f=f+'[%2.2f] ' % q.weighting(userid)
        f=f+q.feedback(userid)+'</li>'
    return f+'</ol>\n'

  def raw_mark(self,userid=None):
    """Calculate the raw mark for given user (ignoring timelimit)"""
    if userid is None: userid=self.getUserid()
    if (not self._submissions.has_key(userid) or
        self._submissions[userid]['submitted'] is None):
      return 0.0
    qs=self.getQuestions(userid)
    sum=0
    ww=0
    for q in qs:
      ww=ww+q.weighting(userid)
      sum=sum+q.weighting(userid)*q.mark(userid)
    return sum/ww

  def mark(self,userid=None):
    """Calculate the mark for given user (including timelimit adjustment)"""
    if userid is None: userid=self.getUserid()
    if self.assessmentStatus(userid): return 0.0
    rmark=self.raw_mark(userid)
    if self.timetaken(userid)<=self.getSetting('timelimit'): return rmark
    return rmark*self.getSetting('timelimit')/self.timetaken(userid)

  def distance_metric(self,id1,id2):
    """measure of difference between submissions for these ids 0.0-1.0"""
    qs=self.getQuestions(id1)
    sum=0
    ww=0
    for q in qs:
      ww=ww+q.weighting()
      sum=sum+q.weighting()*q.distance_metric(id1,id2)
    return sum/ww

  def average(self,values):
    if len(values)==0: return 0
    sum=0
    for v in values: sum=sum+v
    return sum/len(values)

  def histogram(self,values=None,bins=None):
    if bins is None:
      bins=[]
      for i in range(10): bins.append(i*0.1)
    else:
      bins.sort()
    if values is None:
      values=[]
      for i in self.assessedIds():
        values.append(self.mark(i))
    gram=[]
    for i in range(len(bins)): gram.append(0)
    max=0
    for v in values:
      if v>bins[-1]:
        gram[-1]=gram[-1]+1
      else:
        for i in range(len(bins)):
          if v<bins[i+1]:
            gram[i]=gram[i]+1
            if gram[i]>max: max=gram[i]
            break
    h=[]
    for i in range(len(gram)):
      h.append((bins[i],gram[i],))
    return h  
  
  def average_mark(self,userids=None):
    """average mark for given student list, assessed students by default"""
    if userids is None: userids=self.assessedIds()
    return self.average(map (lambda u,self=self: self.mark(u),userids))
    sum=0
    for userid in userids:
      sum=sum+self.mark(userid)
    return sum/len(userids)

  def average_time(self,userids=None):
    """average time taken for given student list,
    assessed students by default"""
    if userids is None: userids=self.assessedIds()
    return self.average(map (lambda u,self=self: self.timetaken(u),userids))
    sum=0
    for userid in userids:
      sum=sum+self.timetaken(userid)
    return sum/len(userids)
   
  def question_statistics(self,userid=None,userids=None):
      """return feedback for given student, throws exception if not allowed"""
      if userid is None: userid=self.getUserid()
      f='<ol>'
      for q in self.all_questions():
          f=f+('\n<li><p><strong>'+q.title_or_id()+
               '</strong></p>'+q.statistics(userid,userids))
      return f+'</ol>\n'

  index_html=HTMLFile('examIndex',globals())

  exam_html=HTMLFile('exam',globals())

  feedback_html=HTMLFile('feedback',globals())

  submit_html=HTMLFile('submit',globals())

  report_html=HTMLFile('report',globals())

  counter=HTMLFile('counter',globals())

  def summary_details(self,userid=None,feedback=0):
    """Useful reporting table for the start of feedback or exam"""
    if userid is None: userid=self.getUserid()
    sub=self.getSubmissionRecord(userid)
    html=("<table border=\"1\" cellspacing=\"0\"><tr><th colspan=\"2\">"+
    self.title_or_id())
    if feedback:
      html=html+" Feedback"
    else:
      html=html+" Exam"
    html=html+"</th></tr>"
    html=html+row('Userid',userid)
    if feedback:
      html=html+row('Your Mark','%2d ' % (self.mark(userid)*100.0))
      html=html+row('The Average Mark','%2d ' % (self.average_mark()*100.0))
      html=html+row('Your Time Taken','%.2d minutes' % self.timetaken(userid))
      html=html+row('Average Time Taken','%.2d minutes' % self.average_time())
    else:
      html=html+row('Form Accessed',DateTime())
    if sub:
      html=html+row('Started',sub['started'])
    else:
      html=html+row('Started',DateTime())
    html=html+row('Timelimit',self.getSetting('timelimit','None'))
    html=html+row('No Questions',self.noQuestions(userid))
    html=html+row('Exam Deadline',self.getSetting('Deadline','None'))
    if sub: html=html+row('Your Deadline',sub['deadline'])
    if sub and sub['submitted']:
      html=html+row('Submission Saved',sub['submitted'])
    else:
      html=html+row('Submission Saved','No')
    stat=self.assessmentStatus(userid)
    if stat:
      html=html+'<tr><td colspan=2 align=left>Assessment not counted due to %s' % stat[1]
    if self.getSetting('randomise_questions'):
      html=html+'<tr><td colspan=2>Questions are Randomised</td></tr>'
    return html+'</table>'

  def collusion_set(self,userid,metric,ids=None):
    """Return list of ids of submissions inside metric
    and overlapping in time"""
    if not self._submissions.has_key(userid): return
    sub1=self._submissions[userid]
    if not sub1: return
    if not ids: ids=self.submittedIds()
    list=[]
    for id in ids:
      if (id<>userid and self._submissions.has_key(id)
          and self.distance_metric(id,userid)<metric):
        sub2=self._submissions[id]
        if ((sub1['started']>=sub2['started']
            and sub1['started']<=sub2['submitted'])
            or
            (sub1['submitted']>=sub2['started']
            and sub1['submitted']<=sub2['submitted'])):
          list.append(id)
    return list
      
def row(label,value,align='left'):
    return ("<tr><th align=\"left\">"+label+
            "</td><td align=\""+align+"\">"+str(value)+"</td></tr>\n")
      
Globals.default__class_init__(Exam)
