__doc__ = "More Exam Questions Classes"
__version__ = "0.0.1"

from OFS.Folder import Folder
from Globals import PersistentMapping,HTMLFile
import Products
from whrandom import randint
from ExamQuestion import ExamMCQ

class ExamMAQ(ExamMCQ):
  """Exam Multiple Answer Question Class"""

  meta_type = "Exam MAQ"
  icon="misc_/Exam/maq"
  
  default_settings= {
    'randomise_answers':1,
    'weight':1.0,
    'init_question_data':None
    }
  
  def __init__(self, id):
    ExamMCQ.__init__(self,id)
    #change C to tokens
    self.manage_delProperties(ids=('C'))
    self.manage_addProperty('C',(1),'tokens')

  def mark(self,userid):
    sub=self.getSubmission(userid)
    if not sub: return 0
    c=self.getSetting('C',userid)
    m=0
    for i in range(len(self.getSetting('A',userid))):
      s=str(i+1)
      if s in sub:
        if s in c: m=m+1
      else:
        if not s in c: m=m+1
    if m==0: return m 
    return m/len(self.getSetting('A',userid))

  def _answer_html(self,text,value,checked=None,disabled=None):
    """generate the entry for an answer"""
    html='<input type="checkbox" name="'+self.id+':list" value="'+str(value)+'"'
    if disabled:
      html=html+' disabled="1"'
    if checked:
      html=html+' checked="1"'
    return html+' > '+str(text)+'\n'

  def question(self,userid,disabled=0):
    """html representation of the question form for given user"""
    self._initSubmission(userid)
    html=self.getSetting('Q',userid)+"\n<p>"
    c=self.getSubmission(userid) or ()
    a=self.getSetting('A',userid)
    for i in self._choiceorder(a):
      html=html+self._answer_html(a[i],i+1,str(i+1) in c,disabled)+'<br />'
    html=html+'</p>\n'
    return html

  def feedback(self,userid):
    """html representation of the feedback for given user"""
    html=self.question(userid,1)+'<p><strong>Your mark '
    html=html+str(self.mark(userid))+' correct choices are '
    for i in self.getSetting('C',userid):
      html=html+'"'+str(self.getSetting('A',userid)[int(i)-1])+'" '
    html=html+'</strong></p>\n'+self.getSetting('E',userid) +'\n'
    return html

  def statistics(self,userid,userids):
    """present statistics for this question - distribution of submissions"""
    submissions=self.getSubmissions(userids)
    html=(self.getSetting('Q',userid)+
          "\n<table border='1' cellspacing='0' cellpadding='0'>\n")
    a=self.getSetting('A',userid)
    c=self.getSetting('C',userid)
    html=html+('<tr><td width="50">'+str(len(filter( lambda u: u==0  ,submissions)))
               +'</td><td>Not Known</td></tr>')
    for i in range(len(a)):
      count=len(filter( lambda u,v=str(i+1): v in u ,submissions))
      html=html+'<tr><td>'+str(count)+'</td><td>'
      if i+1==c:
        html=html+'<strong>'+str(a[i])+'</strong>'
      else:
        html=html+str(a[i])
      html=html+'</td></tr>\n'
    return html+'</table>\n'+self.getSetting('E',userid) +'\n'

  def distance_metric(self,id1,id2):
    """0 of both have submitted the same answer, else 1"""
    sub1=self.getSubmission(id1)
    sub2=self.getSubmission(id2)
    if (sub1 is not None and sub2 is not None and sub1==sub2): return 0.0
    return 1.0
